from django.urls import reverse
from django.test import TestCase

from example_project.polls.factories import UserFactory
from example_project.polls.models import Choice


class LoggedInTestCase(TestCase):
    def setUp(self):
        super(LoggedInTestCase, self).setUp()
        UserFactory.create(
            is_staff=True, is_superuser=True, username="admin", password="admin"
        )
        self.assertTrue(self.client.login(username="admin", password="admin"))


# TODO move most of these to test_admin.py after I sit down and re-read these
# and don't need the fixtures
class AppTests(LoggedInTestCase):
    fixtures = ["sample_data"]

    def test_tool_func_gets_executed(self):
        c = Choice.objects.get(pk=1)
        votes = c.votes
        response = self.client.post(
            reverse("admin:polls_choice_actions", args=(1, "increment_vote"))
        )
        self.assertEqual(response.status_code, 302)
        url = reverse("admin:polls_choice_change", args=(1,))
        self.assertTrue(response["location"].endswith(url))
        c = Choice.objects.get(pk=1)
        self.assertEqual(c.votes, votes + 1)

    def test_tool_func_gets_executed_with_form_params(self):
        c = Choice.objects.get(pk=1)
        votes = c.votes
        response = self.client.post(
            reverse("admin:polls_choice_actions", args=(1, "change_votes")),
            data={"change_by": "10"},
        )
        self.assertEqual(response.status_code, 302)
        url = reverse("admin:polls_choice_change", args=(1,))
        self.assertTrue(response["location"].endswith(url))
        c = Choice.objects.get(pk=1)
        self.assertEqual(c.votes, votes + 10)

    def test_tool_can_return_httpresponse(self):
        # we know this url works because of fixtures
        url = reverse("admin:polls_choice_actions", args=(2, "edit_poll"))
        response = self.client.post(url)
        # we expect a redirect
        self.assertEqual(response.status_code, 302)
        self.assertTrue(
            response["location"].endswith(reverse("admin:polls_poll_change", args=(1,)))
        )

    def test_can_return_template(self):
        # This is more of a test of render_to_response than the app, but I think
        # it's good to document that this is something we can do.
        url = reverse("admin:polls_poll_actions", args=(1, "delete_all_choices"))
        response = self.client.post(url)
        self.assertTemplateUsed(response, "clear_choices.html")

    def test_message_user_sends_message(self):
        url = reverse("admin:polls_poll_actions", args=(1, "delete_all_choices"))
        self.assertNotIn("messages", self.client.cookies)
        self.client.post(url)
        self.assertIn("messages", self.client.cookies)

    def test_intermediate_page_with_post_works(self):
        self.assertTrue(Choice.objects.filter(poll=1).count())
        url = reverse("admin:polls_poll_actions", args=(1, "delete_all_choices"))
        response = self.client.post(url, data={"sure": "Sure"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Choice.objects.filter(poll=1).count(), 0)

    def test_undefined_tool_404s(self):
        response = self.client.post(
            reverse("admin:polls_poll_actions", args=(1, "weeeewoooooo"))
        )
        self.assertEqual(response.status_code, 404)

    def test_key_error_tool_500s(self):
        self.assertRaises(
            KeyError,
            self.client.post,
            reverse("admin:polls_choice_actions", args=(1, "raise_key_error")),
        )

    def test_render_button(self):
        response = self.client.post(reverse("admin:polls_choice_change", args=(1,)))
        self.assertEqual(response.status_code, 200)



"""A Django app for adding object tools for models in the admin."""
__version__ = "5.1.0"


# kind of like __all__, make these available for public
from .utils import (
    BaseDjangoObjectActions,
    DjangoObjectActions,
    takes_instance_or_queryset,
    action,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_object_actions', 'django_object_actions.tests']

package_data = \
{'': ['*'],
 'django_object_actions': ['static/django_object_actions/css/*',
                           'templates/django_object_actions/*']}

setup_kwargs = {
    'name': 'skytek-django-object-actions',
    'version': '5.1.0',
    'description': 'A Django app for adding object tools for models in the admin',
    'long_description': 'Django Object Actions\n=====================\n\n[![CI](https://github.com/crccheck/django-object-actions/actions/workflows/ci.yml/badge.svg?branch=master)](https://github.com/crccheck/django-object-actions/actions/workflows/ci.yml?query=branch%3Amaster)\n\nIf you\'ve ever tried making admin object tools you may have thought, "why can\'t\nthis be as easy as making Django Admin Actions?" Well now they can be.\n\n\nQuick-Start Guide\n-----------------\n\nInstall Django Object Actions:\n\n```shell\n$ pip install django-object-actions\n```\n\nAdd `django_object_actions` to your `INSTALLED_APPS` so Django can find\nour templates.\n\nIn your admin.py:\n\n```python\nfrom django_object_actions import DjangoObjectActions, action\n\nclass ArticleAdmin(DjangoObjectActions, admin.ModelAdmin):\n    @action(label="Publish", description="Submit this article") # optional\n    def publish_this(self, request, obj):\n        publish_obj(obj)\n\n    change_actions = (\'publish_this\', )\n```\n\n\nUsage\n-----\n\nDefining new &*tool actions* is just like defining regular [admin actions]. The\nmajor difference is the functions for `django-object-actions` will take an\nobject instance instead of a queryset (see *Re-using Admin Actions* below).\n\n*Tool actions* are exposed by putting them in a `change_actions` attribute in\nyour `admin.ModelAdmin`. You can also add *tool actions* to the main changelist\nviews too. There, you\'ll get a queryset like a regular [admin action][admin actions]:\n\n```python\nfrom django_object_actions import DjangoObjectActions\n\nclass MyModelAdmin(DjangoObjectActions, admin.ModelAdmin):\n    @action(\n        label="This will be the label of the button",  # optional\n        description="This will be the tooltip of the button" # optional\n    )\n    def toolfunc(self, request, obj):\n        pass\n\n    def make_published(modeladmin, request, queryset):\n        queryset.update(status=\'p\')\n\n    change_actions = (\'toolfunc\', )\n    changelist_actions = (\'make_published\', )\n```\n\nJust like admin actions, you can send a message with `self.message_user`.\nNormally, you would do something to the object and return to the same url, but\nif you return a `HttpResponse`, it will follow it (hey, just like [admin\nactions]!).\n\nIf your admin modifies `get_urls`, `change_view`, or `changelist_view`,\nyou\'ll need to take extra care because `django-object-actions` uses them too.\n\n### Re-using Admin Actions\n\nIf you would like a preexisting admin action to also be an *object action*, add\nthe `takes_instance_or_queryset` decorator to convert object instances into a\nqueryset and pass querysets:\n\n```python\nfrom django_object_actions import DjangoObjectActions, takes_instance_or_queryset\n\nclass RobotAdmin(DjangoObjectActions, admin.ModelAdmin):\n    # ... snip ...\n\n    @takes_instance_or_queryset\n    def tighten_lug_nuts(self, request, queryset):\n        queryset.update(lugnuts=F(\'lugnuts\') - 1)\n\n    change_actions = [\'tighten_lug_nuts\']\n    actions = [\'tighten_lug_nuts\']\n```\n\n[admin actions]: https://docs.djangoproject.com/en/stable/ref/contrib/admin/actions/\n\n### Customizing *Object Actions*\n\nTo give the action some a helpful title tooltip, you can use the `action` decorator\nand set the description argument.\n\n```python\n@action(description="Increment the vote count by one")\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\n```\n\nAlternatively, you can also add a `short_description` attribute,\nsimilar to how admin actions work:\n\n```python\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\nincrement_vote.short_description = "Increment the vote count by one"\n```\n\nBy default, Django Object Actions will guess what to label the button\nbased on the name of the function. You can override this with a `label`\nattribute:\n\n```python\n@action(label="Vote++")\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\n```\n\nor\n\n```python\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\nincrement_vote.label = "Vote++"\n```\n\nIf you need even more control, you can add arbitrary attributes to the buttons\nby adding a Django widget style\n[attrs](https://docs.djangoproject.com/en/stable/ref/forms/widgets/#django.forms.Widget.attrs)\nattribute:\n\n```python\n@action(attrs = {\'class\': \'addlink\'})\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\n```\n\nor\n\n```python\ndef increment_vote(self, request, obj):\n    obj.votes = obj.votes + 1\n    obj.save()\nincrement_vote.attrs = {\n    \'class\': \'addlink\',\n}\n```\n\n## Adding inline forms\n\nYou can add parameters to the action button by adding Django [Form](https://docs.djangoproject.com/en/4.1/ref/forms/api/#django.forms.Form) object to it. Parameter values can be read form request\'s `POST` property.\n\n```python\nfrom django import forms\n\nclass ResetAllForm(forms.Form):\n    new_value = forms.IntegerField(initial=0)\n\ndef reset_all(self, request, queryset):\n    new_value = int(request.POST["new_value"])\n    queryset.update(value=new_value)\nreset_all.form = ResetAllForm()\n```\n\nEach action with form assigned is rendered in it\'s own, separate row.\n\nAlternatively you can set `form_in_modal=True` property, which will open a form on a separate page. This is recommended for complex forms.\n\n\n### Programmatically Disabling Actions\n\nYou can programmatically disable registered actions by defining your own\ncustom `get_change_actions()` method. In this example, certain actions\nonly apply to certain object states (e.g. You should not be able to\nclose an company account if the account is already closed):\n\n```python\ndef get_change_actions(self, request, object_id, form_url):\n    actions = super(PollAdmin, self).get_change_actions(request, object_id, form_url)\n    actions = list(actions)\n    if not request.user.is_superuser:\n        return []\n\n    obj = self.model.objects.get(pk=object_id)\n    if obj.question.endswith(\'?\'):\n        actions.remove(\'question_mark\')\n\n    return actions\n```\n\nThe same is true for changelist actions with `get_changelist_actions`.\n\n### Alternate Installation\n\nYou don\'t have to add this to `INSTALLED_APPS`, all you need to to do\nis copy the template `django_object_actions/change_form.html` some place\nDjango\'s template loader [will find\nit](https://docs.djangoproject.com/en/stable/ref/settings/#template-dirs).\n\nIf you don\'t intend to use the template customizations at all, don\'t\nadd `django_object_actions` to your `INSTALLED_APPS` at all and use\n`BaseDjangoObjectActions` instead of `DjangoObjectActions`.\n\n\nMore Examples\n-------------\n\nMaking an action that links off-site:\n\n```python\ndef external_link(self, request, obj):\n    from django.http import HttpResponseRedirect\n    return HttpResponseRedirect(f\'https://example.com/{obj.id}\')\n```\n\n\nLimitations\n-----------\n\n1.  `django-object-actions` expects functions to be methods of the model\n    admin. While Django gives you a lot more options for their admin\n    actions.\n2.  If you provide your own custom `change_form.html`, you\'ll also need\n    to manually copy in the relevant bits of [our change form\n    ](./django_object_actions/templates/django_object_actions/change_form.html).\n3.  Security. This has been written with the assumption that everyone in\n    the Django admin belongs there. Permissions should be enforced in\n    your own actions irregardless of what this provides. Better default\n    security is planned for the future.\n\n\nPython and Django compatibility\n-------------------------------\n\nSee [`ci.yml`](./.github/workflows/ci.yml) for which Python and Django versions this supports.\n\n\nDemo Admin & Docker images\n--------------------------\n\nYou can try the demo admin against several versions of Django with these Docker\nimages: https://hub.docker.com/r/crccheck/django-object-actions/tags\n\nThis runs the example Django project in `./example_project` based on the "polls"\ntutorial. `admin.py` demos what you can do with this app.\n\n\nDevelopment\n-----------\n\nGetting started:\n\n```shell\n# get a copy of the code\ngit clone git@github.com:crccheck/django-object-actions.git\ncd django-object-actions\n# Install requirements\nmake install\nmake test  # run test suite\nmake quickstart  # runs \'make resetdb\' and some extra steps\n```\n\nVarious helpers are available as make commands. Type `make help` and\nview the `Makefile` to see what other things you can do.\n\nSome commands assume you are in the virtualenv. If you see\n"ModuleNotFoundError"s, try running `poetry shell` first.\n\n\nSimilar Packages\n----------------\n\nIf you want an actions menu for each row of your changelist, check out [Django\nAdmin Row Actions](https://github.com/DjangoAdminHackers/django-admin-row-actions).\n\nDjango Object Actions is very similar to\n[django-object-tools](https://github.com/praekelt/django-object-tools), but does\nnot require messing with your urls.py, does not do anything special with\npermissions, and uses the same patterns as making [admin actions].\n',
    'author': 'crccheck',
    'author_email': 'c@crccheck.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Skytek/django-object-actions',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Integration tests that actually try and use the tools setup in admin.py
"""
from django.contrib.admin.utils import quote
from django.http import HttpResponse
from django.urls import reverse
from unittest.mock import patch

from .tests import LoggedInTestCase
from example_project.polls.factories import (
    ChoiceFactory,
    CommentFactory,
    PollFactory,
    RelatedDataFactory,
)


class CommentTests(LoggedInTestCase):
    def test_action_on_a_model_with_uuid_pk_works(self):
        comment = CommentFactory()
        comment_url = reverse("admin:polls_comment_change", args=(comment.pk,))
        action_url = "/admin/polls/comment/{0}/actions/hodor/".format(comment.pk)
        # sanity check that url has a uuid
        self.assertIn("-", action_url)
        response = self.client.post(action_url)
        self.assertRedirects(response, comment_url)

    @patch("django_object_actions.utils.ChangeActionView.post")
    def test_action_on_a_model_with_arbitrary_pk_works(self, mock_view):
        mock_view.return_value = HttpResponse()
        action_url = "/admin/polls/comment/{0}/actions/hodor/".format(" i am a pk ")

        self.client.post(action_url)

        self.assertTrue(mock_view.called)
        self.assertEqual(mock_view.call_args[1]["pk"], " i am a pk ")

    @patch("django_object_actions.utils.ChangeActionView.post")
    def test_action_on_a_model_with_slash_in_pk_works(self, mock_view):
        mock_view.return_value = HttpResponse()
        action_url = "/admin/polls/comment/{0}/actions/hodor/".format("pk/slash")

        self.client.post(action_url)

        self.assertTrue(mock_view.called)
        self.assertEqual(mock_view.call_args[1]["pk"], "pk/slash")


class ExtraTests(LoggedInTestCase):
    def test_action_on_a_model_with_complex_id(self):
        related_data = RelatedDataFactory()
        related_data_url = reverse(
            "admin:polls_relateddata_change", args=(related_data.pk,)
        )
        action_url = "/admin/polls/relateddata/{}/actions/fill_up/".format(
            quote(related_data.pk)
        )

        response = self.client.post(action_url)
        self.assertNotEqual(response.status_code, 404)
        self.assertRedirects(response, related_data_url)


class ChangeTests(LoggedInTestCase):
    def test_buttons_load(self):
        url = "/admin/polls/choice/"
        response = self.client.get(url)
        self.assertIn("objectactions", response.context_data)
        self.assertIn("Delete all", response.rendered_content)

    def test_changelist_template_context(self):
        url = reverse("admin:polls_poll_changelist")
        response = self.client.get(url)
        self.assertIn("objectactions", response.context_data)
        self.assertIn("tools_view_name", response.context_data)
        self.assertIn("foo", response.context_data)

    def test_changelist_action_view(self):
        url = "/admin/polls/choice/actions/delete_all/"
        response = self.client.post(url)
        self.assertRedirects(response, "/admin/polls/choice/")

    def test_changelist_nonexistent_action(self):
        url = "/admin/polls/choice/actions/xyzzy/"
        response = self.client.post(url)
        self.assertEqual(response.status_code, 404)

    def test_get_changelist_can_remove_action(self):
        poll = PollFactory.create()
        self.assertFalse(poll.question.endswith("?"))
        admin_change_url = reverse("admin:polls_poll_change", args=(poll.pk,))
        action_url = "/admin/polls/poll/1/actions/question_mark/"

        # button is in the admin
        response = self.client.get(admin_change_url)
        self.assertIn(action_url, response.rendered_content)

        response = self.client.post(action_url)  # Click on the button
        self.assertRedirects(response, admin_change_url)

        # button is not in the admin anymore
        response = self.client.get(admin_change_url)
        self.assertNotIn(action_url, response.rendered_content)

    def test_changelist_get_method_action_view(self):
        url = "/admin/polls/choice/actions/delete_all/"
        response = self.client.get(url)
        self.assertEqual(response.status_code, 405)

    def test_changelist_get_method_nonexistent_action(self):
        url = "/admin/polls/choice/actions/xyzzy/"
        response = self.client.get(url)
        self.assertEqual(response.status_code, 405)


class ChangeListTests(LoggedInTestCase):
    def test_changelist_template_context(self):
        poll = PollFactory()
        url = reverse("admin:polls_poll_change", args=(poll.pk,))

        response = self.client.get(url)
        self.assertIn("objectactions", response.context_data)
        self.assertIn("tools_view_name", response.context_data)
        self.assertIn("foo", response.context_data)


class MultipleAdminsTests(LoggedInTestCase):
    def test_redirect_back_from_secondary_admin(self):
        poll = PollFactory()
        admin_change_url = reverse(
            "admin:polls_poll_change", args=(poll.pk,), current_app="support"
        )
        action_url = "/support/polls/poll/1/actions/question_mark/"
        self.assertTrue(admin_change_url.startswith("/support/"))

        response = self.client.post(action_url)
        self.assertRedirects(response, admin_change_url)


class FormTests(LoggedInTestCase):
    def test_form_is_rendered_in_change_view(self):
        choice = ChoiceFactory()
        admin_change_url = reverse("admin:polls_choice_change", args=(choice.pk,))

        response = self.client.get(admin_change_url)

        # form is in the admin
        action_url_lookup = 'action="/admin/polls/choice/1/actions/change_votes/"'
        self.assertIn(action_url_lookup, response.rendered_content)
        form_lookup = '<form name="change_votes__form"'
        self.assertIn(form_lookup, response.rendered_content)

        # form has input
        input_lookup = 'name="change_by"'
        self.assertIn(input_lookup, response.rendered_content)

    def test_form_is_rendered_in_changelist(self):
        admin_change_url = reverse("admin:polls_choice_changelist")

        response = self.client.get(admin_change_url)

        # form is in the admin
        action_url_lookup = 'action="/admin/polls/choice/actions/reset_all/"'
        self.assertIn(action_url_lookup, response.rendered_content)
        form_lookup = '<form name="reset_all__form"'
        self.assertIn(form_lookup, response.rendered_content)

        # form has input
        input_lookup = 'name="new_value"'
        self.assertIn(input_lookup, response.rendered_content)
